from __future__ import annotations

import functools
import operator
import string
import types
from typing import *

from v440.core.VersionError import VersionError


class Digest:
    __slots__ = ("__dict__", "lookup", "name", "kind")
    lookup: dict
    name: str
    kind: Any

    def __call__(self: Self, *args: Any, **kwargs: Any) -> Any:
        "This magic method implements self(*args, **kwargs)."
        return self.wrapped(*args, **kwargs)

    def __get__(
        self: Self,
        *args: Any,
        **kwargs: Any,
    ) -> types.FunctionType | types.MethodType:
        "This magic method implements getting as an attribute from a class or an object."
        return self.wrapped.__get__(*args, **kwargs)

    def __init__(
        self: Self,
        name: str = "",
        kind: Any = None,
    ) -> None:
        "This magic method sets up self."
        self.lookup = dict()
        self.name = name
        self.kind = kind

    @classmethod
    def _getkey(cls: type, value: Any) -> Any:
        if value is None:
            return
        if isinstance(value, int):
            return int
        if isinstance(value, str):
            return str
        try:
            value.__iter__
        except AttributeError:
            return str
        else:
            return list

    def _overload(self: Self, key: Any, value: Any) -> Self:
        self.lookup[key] = value
        overload(value)
        return self

    def overload(self: Self, key: Any = None) -> functools.partial:
        return functools.partial(type(self)._overload, self, key)

    @functools.cached_property
    def wrapped(self: Self) -> Any:
        def new(*args: Any, **kwargs: Any) -> Any:
            args0: list = list(args)
            value: Any = args0.pop()
            key: Any = self._getkey(value)
            if key is int:
                args0.append(int(value))
            if key is str:
                args0.append(str(value).lower().strip())
            if key is list:
                args0.append(list(value))
            ans: Any = self.lookup[key](*args0, **kwargs)
            return ans

        new.__name__ = self.name
        if self.kind is not None:
            new = self.kind(new)
        return new


def literal(value: Any, /) -> str:
    v: Any = segment(value)
    if type(v) is str:
        return v
    e: str = "%r is not a valid literal segment"
    e %= v
    raise VersionError(e)


def numeral(value: Any, /) -> int:
    v: Any = segment(value)
    if type(v) is int:
        return v
    e = "%r is not a valid numeral segment"
    e %= v
    raise VersionError(e)


def segment(value: Any, /) -> Any:
    try:
        return _segment(value)
    except:
        e = "%r is not a valid segment"
        e = VersionError(e % value)
        raise e from None


_segment: Digest = Digest("_segment")


@_segment.overload()
def _segment():
    return


@_segment.overload(int)
def byInt(value: int, /) -> Any:
    if value < 0:
        raise ValueError
    return value


@_segment.overload(str)
def _segment(value: Any, /) -> int | str:
    if value.strip(string.ascii_lowercase + string.digits):
        raise ValueError(value)
    if value.strip(string.digits):
        return value
    elif value == "":
        return 0
    else:
        return int(value)


def torange(key: Any, length: Any) -> range:
    start: Any = key.start
    stop: Any = key.stop
    step: Any = key.step
    if step is None:
        step = 1
    else:
        step = operator.index(step)
        if step == 0:
            raise ValueError
    fwd: bool = step > 0
    if start is None:
        start = 0 if fwd else (length - 1)
    else:
        start = operator.index(start)
    if stop is None:
        stop = length if fwd else -1
    else:
        stop = operator.index(stop)
    if start < 0:
        start += length
    if start < 0:
        start = 0 if fwd else -1
    if stop < 0:
        stop += length
    if stop < 0:
        stop = 0 if fwd else -1
    ans: range = range(start, stop, step)
    return ans



from __future__ import annotations

import dataclasses
from typing import *

from v440._utils import utils
from v440._utils.Pattern import Pattern


class QualifierParser: ...


@dataclasses.dataclass(frozen=True)
class SimpleQualifierParser(QualifierParser):
    keysforlist: tuple = ()
    keysforstr: tuple = ()
    allow_len_1: bool = False

    __call__ = utils.Digest("__call__")

    @__call__.overload()
    def __call__(self: Self) -> Optional[list]:
        pass

    @__call__.overload(int)
    def __call__(self: Self, value: int) -> Any:
        if value < 0:
            raise ValueError
        return value

    @__call__.overload(list)
    def __call__(self: Self, value: list) -> Any:
        v: list = list(map(utils.segment, value))
        n = self.nbylist(v)
        if isinstance(n, str):
            raise TypeError
        return n

    @__call__.overload(str)
    def __call__(self: Self, value: str) -> Optional[int | list]:
        v: str = value
        v = v.replace("_", ".")
        v = v.replace("-", ".")
        m: Any = Pattern.PARSER.bound.search(v)
        l, n = m.groups()
        if l not in self.keysforstr:
            raise ValueError
        if n is None:
            return None
        else:
            return int(n)

    def __post_init__(self: Self) -> None:
        if type(self.allow_len_1) is not bool:
            raise TypeError
        if type(self.keysforlist) is not tuple:
            raise TypeError
        if type(self.keysforstr) is not tuple:
            raise TypeError
        keys = self.keysforlist + self.keysforstr
        for k in keys:
            if k is None:
                continue
            if type(k) is str:
                continue
            raise TypeError

    def nbylist(self: Self, value: Any, /) -> Any:
        if len(value) == 2:
            l, n = value
            if l not in self.keysforlist:
                raise ValueError
            return n
        if len(value) == 1:
            n = value[0]
            if not self.allow_len_1:
                raise ValueError
            return n
        raise ValueError


class PhasedQualifierParser(QualifierParser):
    __slots__ = ("_phasedict",)

    phasedict: dict

    __call__ = utils.Digest("__call__")

    @__call__.overload()
    def __call__(self: Self) -> list:
        return [None, None]

    @__call__.overload(list)
    def __call__(self: Self, value: list) -> Any:
        l: Any
        n: Any
        l, n = list(map(utils.segment, value))
        if [l, n] == [None, None]:
            return [None, None]
        l = self.phasedict[l]
        if not isinstance(n, int):
            raise TypeError
        return [l, n]

    @__call__.overload(str)
    def __call__(self: Self, value: str) -> list:
        if value == "":
            return [None, None]
        v: str = value
        v = v.replace("_", ".")
        v = v.replace("-", ".")
        m: Any = Pattern.PARSER.bound.search(v)
        l: Any
        n: Any
        l, n = m.groups()
        l = self.phasedict[l]
        n = 0 if (n is None) else int(n)
        return [l, n]

    def __init__(self: Self, **kwargs: Any) -> None:
        self._phasedict = dict()
        x: Any
        y: Any
        for x, y in kwargs.items():
            self._phasedict[str(x)] = str(y)

    def nbylist(self: Self, value: Any, /) -> Any:
        raise ValueError

    @property
    def phasedict(self: Self) -> dict:
        return dict(self._phasedict)


POST = SimpleQualifierParser(
    keysforlist=("post", "rev", "r", ""),
    keysforstr=(None, "post", "rev", "r"),
    allow_len_1=True,
)
DEV = SimpleQualifierParser(
    keysforlist=("dev",),
    keysforstr=(None, "dev"),
)
PRE = PhasedQualifierParser(
    alpha="a",
    a="a",
    beta="b",
    b="b",
    preview="rc",
    pre="rc",
    c="rc",
    rc="rc",
)



from __future__ import annotations

import functools
from typing import *

from v440._utils import utils
from v440._utils.VList import VList

__all__ = ["Local"]


class Local(VList):

    data: list[int | str]

    def __le__(self: Self, other: Iterable) -> bool:
        "This magic method implements self<=other."
        ans: bool
        try:
            alt: Self = type(self)(other)
        except ValueError:
            ans = self.data <= other
        else:
            ans = self._cmpkey() <= alt._cmpkey()
        return ans

    def __str__(self: Self) -> str:
        "This magic method implements str(self)."
        return ".".join(map(str, self))

    def _cmpkey(self: Self) -> list:
        return list(map(self._sortkey, self))

    _data_calc: utils.Digest = utils.Digest("_data_calc")

    @_data_calc.overload()
    def _data_calc(self: Self) -> None:
        return list()

    @_data_calc.overload(int)
    def _data_calc(self: Self, value: int) -> list:
        return [value]

    @_data_calc.overload(list)
    def _data_calc(self: Self, value: list) -> list:
        ans: list = list(map(utils.segment, value))
        if None in ans:
            raise ValueError
        return ans

    @_data_calc.overload(str)
    def _data_calc(self: Self, value: str) -> None:
        v: str = value
        if v.startswith("+"):
            v = v[1:]
        v = v.replace("_", ".")
        v = v.replace("-", ".")
        ans: list = v.split(".")
        ans = list(map(utils.segment, ans))
        if None in ans:
            raise ValueError
        return ans

    @staticmethod
    def _sortkey(value: Any) -> Tuple[bool, Any]:
        return type(value) is int, value

    @property
    def data(self: Self) -> list[int | str]:
        return list(self._data)

    @data.setter
    def data(self: Self, value: Any) -> None:
        self._data = self._data_calc(value)

    @functools.wraps(VList.sort)
    def sort(self: Self, /, *, key: Any = None, **kwargs: Any) -> None:
        k: Any = self._sortkey if key is None else key
        self._data.sort(key=k, **kwargs)



from __future__ import annotations

import operator
import string
from typing import *

from keyalias import keyalias
from overloadable import Overloadable

from v440._utils import utils
from v440._utils.VList import VList


@keyalias(major=0, minor=1, micro=2, patch=2)
class Release(VList):
    data: list[int]
    major: int
    minor: int
    micro: int
    patch: int

    def __add__(self: Self, other: Any, /) -> Self:
        opp: Self = type(self)(other)
        ans: Self = self.copy()
        ans._data += opp._data
        return ans

    @Overloadable
    def __delitem__(self: Self, key: Any) -> bool:
        return type(key) is slice

    @__delitem__.overload(False)
    def __delitem__(self: Self, key: SupportsIndex) -> None:
        i: int = operator.index(key)
        if i < len(self):
            del self._data[i]

    @__delitem__.overload(True)
    def __delitem__(self: Self, key: Any) -> None:
        r: range = utils.torange(key, len(self))
        l: list = [k for k in r if k < len(self)]
        l.sort(reverse=True)
        for k in l:
            del self._data[k]

    @Overloadable
    def __getitem__(self: Self, key: Any) -> bool:
        return type(key) is slice

    @__getitem__.overload(False)
    def __getitem__(self: Self, key: Any) -> int:
        i: int = operator.index(key)
        ans: int = self._getitem_int(i)
        return ans

    @__getitem__.overload(True)
    def __getitem__(self: Self, key: Any) -> list:
        r: range = utils.torange(key, len(self))
        m: map = map(self._getitem_int, r)
        ans: list = list(m)
        return ans

    @Overloadable
    def __setitem__(self: Self, key: Any, value: Any) -> bool:
        return type(key) is slice

    @__setitem__.overload(False)
    def __setitem__(self: Self, key: SupportsIndex, value: Any) -> Any:
        i: int = operator.index(key)
        self._setitem_int(i, value)

    @__setitem__.overload(True)
    def __setitem__(self: Self, key: SupportsIndex, value: Any) -> Any:
        k: range = utils.torange(key, len(self))
        self._setitem_range(k, value)

    def __str__(self: Self) -> str:
        return self.format()

    def _getitem_int(self: Self, key: int) -> int:
        if key < len(self):
            return self._data[key]
        else:
            return 0

    def _setitem_int(self: Self, key: int, value: Any) -> Any:
        v: int = utils.numeral(value)
        n: int = len(self)
        if n > key:
            self._data[key] = v
            return
        if v == 0:
            return
        self._data.extend([0] * (key - n))
        self._data.append(v)

    @Overloadable
    def _setitem_range(self: Self, key: range, value: Any) -> bool:
        return key.step == 1

    @_setitem_range.overload(False)
    def _setitem_range(self: Self, key: range, value: Any) -> Any:
        key = list(key)
        value = self._tolist(value, slicing=len(key))
        if len(key) != len(value):
            e = "attempt to assign sequence of size %s to extended slice of size %s"
            e %= (len(value), len(key))
            raise ValueError(e)
        maximum = max(*key)
        ext = max(0, maximum + 1 - len(self))
        data = self.data
        data += [0] * ext
        for k, v in zip(key, value):
            data[k] = v
        while len(data) and not data[-1]:
            data.pop()
        self._data = data

    @_setitem_range.overload(True)
    def _setitem_range(self: Self, key: range, value: Any) -> Any:
        data: list = self.data
        ext: int = max(0, key.start - len(data))
        data += ext * [0]
        l: list = self._tolist(value, slicing="always")
        data = data[: key.start] + l + data[key.stop :]
        while len(data) and not data[-1]:
            data.pop()
        self._data = data

    @Overloadable
    @staticmethod
    def _tolist(value: Any, *, slicing: Any) -> list:
        if value is None:
            return
        if isinstance(value, int):
            return int
        if hasattr(value, "__iter__") and not isinstance(value, str):
            return list
        return str

    @_tolist.overload()
    def _tolist(value: None, *, slicing: Any) -> list:
        return list()

    @_tolist.overload(int)
    def _tolist(value: int, *, slicing: Any) -> list:
        return [utils.numeral(value)]

    @_tolist.overload(list)
    def _tolist(value: int, *, slicing: Any) -> list:
        return list(map(utils.numeral, value))

    @_tolist.overload(str)
    def _tolist(value: Any, *, slicing: Any) -> list:
        s: Any
        if isinstance(value, str):
            s = slicing
        else:
            s = "never"
        v: str = str(value)
        if v == "":
            return list()
        if "" == v.strip(string.digits) and s in (len(v), "always"):
            return list(map(int, v))
        v = v.lower().strip()
        v = v.replace("_", ".")
        v = v.replace("-", ".")
        if v.startswith("v") or v.startswith("."):
            v = v[1:]
        l: list = v.split(".")
        if "" in l:
            raise ValueError
        l = list(map(utils.numeral, l))
        return l

    def bump(self: Self, index: SupportsIndex = -1, amount: SupportsIndex = 1) -> None:
        i: int = operator.index(index)
        a: int = operator.index(amount)
        x: int = self._getitem_int(i) + a
        self._setitem_int(i, x)
        if i != -1:
            self.data = self.data[: i + 1]

    @property
    def data(self: Self) -> list:
        return list(self._data)

    @data.setter
    def data(self: Self, value: Any) -> None:
        v: list = self._tolist(value, slicing="always")
        while v and v[-1] == 0:
            v.pop()
        self._data = v

    def format(self: Self, cutoff: Any = None) -> str:
        s: str = str(cutoff) if cutoff else ""
        i: Optional[int] = int(s) if s else None
        l: list = self[:i]
        if len(l) == 0:
            l += [0]
        l = list(map(str, l))
        ans: str = ".".join(l)
        return ans
